"""Contains wrappers for yaml.

Each wrapper is used to avoid context manager boilerplate.

By default this module uses ``yaml.FullLoader`` for reading.
See https://msg.pyyaml.org/load.

Attributes:
    read_yaml: Wrapped ``yaml.load``.
    write_yaml: Wrapped ``yaml.dump``.
    read_all_yaml: Wrapped ``yaml.load_all``.
    write_all_yaml: Wrapped ``yaml.dump_all``.

"""

import yaml

from .utils import read_wrapper, write_wrapper


read_yaml = read_wrapper(yaml.load, mode='r', Loader=yaml.FullLoader)
write_yaml = write_wrapper(yaml.dump, mode='w')

read_all_yaml = read_wrapper(yaml.load_all, mode='r', Loader=yaml.FullLoader)
write_all_yaml = write_wrapper(yaml.dump_all, mode='w')

__all__ = ['read_yaml', 'write_yaml',
           'read_all_yaml', 'write_all_yaml']



import dataclasses
import datetime
import logging
from dataclasses import dataclass
from enum import Enum
from typing import Optional, List, Tuple

from dateutil.parser import parse

logger = logging.getLogger(__name__)

from homelypy.states import (
    State,
    BatteryState,
    DiagnosticState,
    TemperatureState,
    BasicAlarmState,
    SmokeAlarmState,
    MotionSensorState,
    MeteringState,
)


class AlarmStates(Enum):
    DISARMED = "DISARMED"
    ARMED_AWAY = "ARMED_AWAY"
    ARMED_NIGHT = "ARMED_NIGHT"
    ARMED_STAY = "ARMED_STAY"
    BREACHED = "BREACHED"
    ALARM_PENDING = "ALARM_PENDING"
    ARM_STAY_PENDING = "ARM_STAY_PENDING"
    ARM_NIGHT_PENDING = "ARM_NIGHT_PENDING"
    ARM_PENDING = "ARM_PENDING"


@dataclass
class Location:
    name: str
    role: str
    user_id: str
    location_id: str
    gateway_serial: str

    def __str__(self):
        return self.name


@dataclass
class Device:
    id: str
    name: str
    serial_number: str
    location: str
    online: bool
    model_id: str
    model_name: str

    def __str__(self):
        return f"{self.name} in {self.location}"

    def update_state(self, changes: List[dict]):
        """Updates the various features of the device based on the list of changes"""
        # Expects the contents of the ["data"]["changes"] key.
        # {
        #     "type": "device-state-changed",
        #     "data": {
        #         "deviceId": "ad5d19b5-3988-4ad2-96c0-08f6283e073a",
        #         "gatewayId": "3b0187f4-878e-4b51-af2b-fc563b81f137",
        #         "locationId": "48617520-863c-4e27-9a05-4ce3cce50f8e",
        #         "modelId": "87fa1ae0-824f-4d42-be7a-cc5b6c7b1e35",
        #         "rootLocationId": "d14a27d8-311c-41d8-b8c1-08b757c2253f",
        #         "changes": [
        #             {
        #                 "feature": "temperature",
        #                 "stateName": "temperature",
        #                 "value": 4.8,
        #                 "lastUpdated": "2023-01-25T10:27:07.786Z",
        #             }
        #         ],
        #     },
        # }
        updated_states = []
        for change in changes:
            try:
                state = getattr(self, change["feature"])
                setattr(state, change["stateName"], change["value"])
                setattr(state, f"{change['stateName']}_last_updated", parse(change["lastUpdated"]))
                updated_states.append(state)
            except AttributeError:
                logger.exception(f"Device '{self}' does not have the feature {change['feature']}")
        return updated_states

    @classmethod
    def create_from_rest_response(cls, device: dict) -> "Device":
        state_fields = {}
        for field in dataclasses.fields(cls):
            if issubclass(field.type, State):
                state_fields[field.name] = field.type.create_from_rest_response(device)
        return cls(
            device["id"],
            device["name"],
            device["serialNumber"],
            device["location"],
            device["online"],
            device["modelId"],
            device["modelName"],
            **state_fields,
        )

    def get_entities(self) -> list[State]:
        return [getattr(self, field.name) for field in dataclasses.fields(self) if issubclass(field.type, State)]


@dataclass
class WindowSensor(Device):
    battery: BatteryState
    diagnostic: DiagnosticState
    temperature: TemperatureState
    alarm: BasicAlarmState


@dataclass
class SmokeAlarm(Device):
    battery: BatteryState
    diagnostic: DiagnosticState
    temperature: TemperatureState
    alarm: SmokeAlarmState


@dataclass
class MotionSensorMini(Device):
    battery: BatteryState
    diagnostic: DiagnosticState
    temperature: TemperatureState
    alarm: MotionSensorState


@dataclass
class EMIHANPowersSensor(Device):
    diagnostic: DiagnosticState
    metering: MeteringState


DEVICE_MAP = {
    "Motion Sensor Mini": MotionSensorMini,
    "Smoke Alarm": SmokeAlarm,
    "Intelligent Smoke Alarm": SmokeAlarm,
    "Window Sensor": WindowSensor,
    "Window Alarm Sensor": WindowSensor,
    "EMI Norwegian HAN": EMIHANPowersSensor,
}


class UnknownDeviceException(Exception):
    pass


def create_device_from_rest_response(data: dict) -> Optional[Device]:
    device_class = DEVICE_MAP.get(data.get("modelName"))
    if device_class is None:
        raise UnknownDeviceException(f"Unknown device: '{data.get('modelName')}'")
    return device_class.create_from_rest_response(data)


@dataclass
class SingleLocation:
    location_id: str
    gateway_serial: str
    name: str
    alarm_state: AlarmStates
    alarm_state_last_updated: datetime.datetime
    user_role_at_location: str
    devices: list[Device]

    def __str__(self):
        return f"{self.name} with {len(self.devices)} devices"

    def find_device(self, device_id) -> Optional[Device]:
        return next(filter(lambda d: d.id == device_id, self.devices), None)

    def update_device_state_from_stream(self, data: dict) -> Optional[Tuple[Device, List[State]]]:
        """
        Updates the device state based on the data package received from the Homely websocket stream. Returns the
        updated device.
        """
        # Expects the contents of the "data" key.
        # {
        #     "type": "device-state-changed",
        #     "data": {
        #         "deviceId": "ad5d19b5-3988-4ad2-96c0-08f6283e073a",
        #         "gatewayId": "3b0187f4-878e-4b51-af2b-fc563b81f137",
        #         "locationId": "48617520-863c-4e27-9a05-4ce3cce50f8e",
        #         "modelId": "87fa1ae0-824f-4d42-be7a-cc5b6c7b1e35",
        #         "rootLocationId": "d14a27d8-311c-41d8-b8c1-08b757c2253f",
        #         "changes": [
        #             {
        #                 "feature": "temperature",
        #                 "stateName": "temperature",
        #                 "value": 4.8,
        #                 "lastUpdated": "2023-01-25T10:27:07.786Z",
        #             }
        #         ],
        #     },
        # }
        device = self.find_device(data["deviceId"])
        if device:
            states = device.update_state(data["changes"])
            return device, states
        else:
            logger.warning(f"Did not find a device matching data update: {data}")
            return None



from .builtins_wrapper import *
from .builtins_wrapper import __all__ as builtins_wrapper_all

try:
    from .yaml_wrapper import *
    from .yaml_wrapper import __all__ as yaml_wrapper_all
except ImportError:
    yaml_wrapper_all = []

from .__version__ import __version__

__all__ = [*builtins_wrapper_all, *yaml_wrapper_all]



"""
Parser for engine.io protocol

This implementation is largely inspired from the package python-engineio
written by Miguel Grinberg and available under the MIT license at
https://github.com/miguelgrinberg/python-engineio
"""

import six

class Packet(object):
    OPEN    = 0
    CLOSE   = 1
    PING    = 2
    PONG    = 3
    MESSAGE = 4
    UPGRADE = 5
    NOOP    = 6

    def __init__(self, _type, data='', binary=False):
        self.type = _type
        self.data = data
        self.binary = binary

    @property
    def type_string(self):
        return {
            self.OPEN:     'open',
            self.CLOSE:    'close',
            self.PING:     'ping',
            self.PONG:     'pong',
            self.MESSAGE:  'message',
            self.UPGRADE:  'upgrade',
            self.NOOP:     'noop'
        }[self.type]

class Parser(object):
    def decode_packet(self, bytes):
        """Decode a received package."""
        b64 = False
        if not isinstance(bytes, six.binary_type):
            bytes = bytes.encode('utf-8')

        packet_type = six.byte2int(bytes[0:1])
        if packet_type == ord('b'):
            binary = True
            bytes = bytes[1:]
            packet_type = int(chr(six.byte2int(bytes[0:1])))
            b64 = True
        elif packet_type >= ord('0'):
            packet_type = int(chr(packet_type))
            binary = False
        else:
            binary = True

        packet_data = None
        if len(bytes) > 1:
            if binary:
                if b64:
                    packet_data = base64.b64decode(bytes[1:])
                else:
                    packet_data = bytes[1:]
            else:
                packet_data = bytes[1:].decode('utf-8')

        return Packet(packet_type, packet_data, binary)

    def encode_packet(self, packet, b64=False):
        """Encode the packet to be sent."""
        if packet.binary and not b64:
            bytes = six.int2byte(packet.type)
        else:
            bytes = six.binary_type(packet.type)
            if packet.binary and b64:
                bytes = 'b' + bytes

        if packet.binary and b64:
            bytes += base64.b64encode(packet.data).decode('utf-8')
        else:
            bytes += six.binary_type(packet.data)

        return bytes

    def decode_payload(self, bytes):
        """Decode a received payload."""
        packets = []
        while bytes:
            if six.byte2int(bytes[0:1]) <= 1:
                packet_len = 0
                i = 1
                while six.byte2int(bytes[i:i + 1]) != 255:
                    packet_len = packet_len * 10 + six.byte2int(bytes[i:i + 1])
                    i += 1
                packet_start = i+1
            else:
                i = bytes.find(b':')
                if i == -1:
                    raise ValueError('Invalid payload')
                packet_len = int(bytes[0:i])
                packet_start = i+1

            packet = self.decode_packet(bytes[packet_start:packet_start+packet_len])
            packets.append(packet)
            bytes = bytes[packet_start+packet_len:]

        return packets

    def encode_payload(self, packets, b64=False):
        """Encode the payload to be sent."""
        bytes = b''
        for packet in packets:
            packet_bytes = self.encode_packet(packet, b64)
            packet_len = len(packet_bytes)
            if b64:
                bytes += str(packet_len) + b':' + packet_bytes
            else:
                binary_len = b''
                while packet_len != 0:
                    binary_len = six.int2byte(packet_len % 10) + binary_len
                    packet_len = int(packet_len / 10)
                bytes += b'\x01' if packet.binary else b'\x00'
                bytes += binary_len + b'\xff' + packet_bytes

        return bytes




__author__ = 'ShortIO Team'
__author_email__ = 'shortio@protonmail.com'
__description__ = 'ShortIO library is intended to avoid context manager boilerplate in simple io operations.'
__license__ = 'MIT'
__title__ = 'shortio'
__url__ = 'https://github.com/byshyk/shortio'
__version__ = '0.1.0'



import datetime
import logging

from dateutil.parser import parse

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(threadName)-15s %(name)-15s: %(levelname)-8s %(message)s",
    datefmt="%d/%m/%Y %H:%M:%S",
)
import argparse
import json
import time
from getpass import getpass
from typing import Callable, Dict, List, Tuple, Any, Optional

import requests
import socketio
import websocket

from homelypy.devices import Location, SingleLocation, create_device_from_rest_response, UnknownDeviceException, Device
from homelypy.states import State

WEB_SOCKET_URL = "wss://sdk.iotiliti.cloud"

SDK_URL = "https://sdk.iotiliti.cloud"
AUTHENTICATION_ENDPOINT = "/homely/oauth/token"
LOCATIONS_ENDPOINT = "/homely/locations"
SINGLE_LOCATION_ENDPOINT = "/homely/home"
REFRESH_TOKEN_ENDPOINT = "/homely/oauth/refresh-token"

logger = logging.getLogger(__name__)


class ConnectionFailedException(Exception):
    pass


class AuthenticationFailedException(Exception):
    pass


class Homely:
    single_location: SingleLocation
    state_change_callback: Callable[[SingleLocation, Device, List[State]], Any]
    sio: socketio.Client

    def __init__(self, username: str, password: str):
        super().__init__()
        self.refresh_expires_in = 0
        self.expires_in = 0
        self.authentication_time = 0
        self.access_token = None
        self.refresh_token = None
        self.username = username
        self.password = password

    def _register_callbacks(self):
        @self.sio.event
        def connect():
            logger.info("websocket: connected to server")

        @self.sio.event
        def disconnect():
            logger.info("websocket: disconnected from server")
            # Disconnected, refresh login
            self.sio.connection_headers = self.build_connection_header()

        @self.sio.on("event")
        def on_message(data):
            # logger.info(data)
            # {
            #     "type": "device-state-changed",
            #     "data": {
            #         "deviceId": "ad5d19b5-3988-4ad2-96c0-08f6283e073a",
            #         "gatewayId": "3b0187f4-878e-4b51-af2b-fc563b81f137",
            #         "locationId": "48617520-863c-4e27-9a05-4ce3cce50f8e",
            #         "modelId": "87fa1ae0-824f-4d42-be7a-cc5b6c7b1e35",
            #         "rootLocationId": "d14a27d8-311c-41d8-b8c1-08b757c2253f",
            #         "changes": [
            #             {
            #                 "feature": "temperature",
            #                 "stateName": "temperature",
            #                 "value": 4.8,
            #                 "lastUpdated": "2023-01-25T10:27:07.786Z",
            #             }
            #         ],
            #     },
            # }
            if data["type"] == "device-state-changed":
                if self.single_location:
                    device, states = self.single_location.update_device_state_from_stream(data["data"])
                    if self.state_change_callback:
                        self.state_change_callback(None, device, states)
            elif data["type"] == "alarm-state-changed":
                if self.single_location:
                    self.single_location.alarm_state = data["data"]["state"]
                    self.single_location.alarm_state_last_updated = parse(data["data"]["timestamp"])
                    if self.state_change_callback:
                        self.state_change_callback(self.single_location, None, [])

    @staticmethod
    def url(endpoint: str) -> str:
        return SDK_URL + endpoint

    def authenticate(self):
        response = requests.post(
            self.url(AUTHENTICATION_ENDPOINT),
            data={"username": self.username, "password": self.password},
        )
        if response.status_code == 401:
            raise AuthenticationFailedException(response.text)
        if response.status_code != 201:
            raise ConnectionFailedException(response.text)
        data = response.json()
        self.store_authentication_information(data)

    def store_authentication_information(self, data: Dict):
        self.access_token = data["access_token"]
        self.authentication_time = time.time()
        self.expires_in = data["expires_in"]
        self.refresh_token = data["refresh_token"]
        self.refresh_expires_in = data["refresh_expires_in"]

    def reauthenticate(self):
        response = requests.post(
            self.url(REFRESH_TOKEN_ENDPOINT),
            data={"refresh_token": self.refresh_token},
        )
        if response.status_code != 201:
            self.refresh_expires_in = -1
            self.expires_in = -1
            raise ConnectionFailedException(response.text)
        data = response.json()
        self.store_authentication_information(data)

    def authenticate_if_required(self):
        if time.time() - self.authentication_time > self.refresh_expires_in - 2:
            self.authenticate()
        elif time.time() - self.authentication_time > self.expires_in - 2:
            self.reauthenticate()

    @property
    def authorisation_header(self) -> Dict:
        self.authenticate_if_required()
        return {"Authorization": f"Bearer {self.access_token}"}

    def get_locations(self) -> List[Location]:
        response = requests.get(self.url(LOCATIONS_ENDPOINT), headers=self.authorisation_header)
        if response.status_code != 200:
            raise ConnectionFailedException(response.text)
        return [
            Location(
                location_data["name"],
                location_data["role"],
                location_data["userId"],
                location_data["locationId"],
                location_data["gatewayserial"],
            )
            for location_data in response.json()
        ]

    def get_location_json(self, location_id) -> dict:
        response = requests.get(
            self.url(SINGLE_LOCATION_ENDPOINT) + f"/{location_id}",
            headers=self.authorisation_header,
        )
        if response.status_code != 200:
            raise ConnectionFailedException(response.text)
        return response.json()

    def get_location(self, location_id) -> SingleLocation:
        data = self.get_location_json(location_id)
        devices = []
        for device in data["devices"]:
            try:
                devices.append(create_device_from_rest_response(device))
            except UnknownDeviceException as ex:
                logger.error(str(ex))
        return SingleLocation(
            data["locationId"],
            data["gatewayserial"],
            data["name"],
            data["alarmState"],
            datetime.datetime.now(datetime.timezone.utc),
            data["userRoleAtLocation"],
            devices,
        )

    def build_connection_header(self) -> dict:
        return {**self.authorisation_header, "locationId": self.single_location.location_id}

    def run_socket_io(
        self,
        single_location: SingleLocation,
        state_change_callback: Optional[Callable[[Device, List[State]], Any]] = None,
    ):
        self.state_change_callback = state_change_callback
        self.single_location = single_location
        self.authenticate_if_required()
        websocket.enableTrace(True)
        url = f"{WEB_SOCKET_URL}?locationId={single_location.location_id}&token=Bearer%20{self.access_token}"
        logger.debug(f"Connecting to web socket {url}")
        self.sio = socketio.Client(logger=logger, engineio_logger=False)
        self._register_callbacks()
        logging.getLogger("socketio").setLevel(logger.level)
        logging.getLogger("websocket").setLevel(logger.level)
        logging.getLogger("engineio").setLevel(logger.level)
        while True:
            try:
                header = self.build_connection_header()
                self.sio.connect(url, headers=header)
                self.sio.wait()
            except:
                logger.exception("Exception while running socketio")
            try:
                self.sio.disconnect()
            except Exception as ex:
                logger.warning(f"Failed disconnecting after unexpected websocket termination: {ex}")
            time.sleep(30)


def test_callback(single_location: Optional[SingleLocation], device: Optional[Device], states: List[State]):
    if device is not None:
        logger.info(f"Received update for device '{device}'")
        for state in states:
            logger.info(f"\t{state.feature_name} change to {state}")
    if single_location is not None:
        logger.info(f"Received alarm update: {single_location.alarm_state}")


if __name__ == "__main__":

    parser = argparse.ArgumentParser(prog="Homelypy", description="Query the Homely rest API")
    parser.add_argument("username", help="Same username as in the Homely app")
    parser.add_argument("-s", "--stream", action="store_true", help="Initiate websocket stream")
    args = parser.parse_args()
    password = getpass()

    homely = Homely(args.username, password)
    locations = homely.get_locations()
    for location in locations:
        logger.info(f"Received location '{location}'")
    for location in locations:
        filename = f"location_{location.name}.json"
        location_dictionary = homely.get_location_json(location.location_id)
        # pprint(location_dictionary)
        # logger.info("----------------------------------")
        with open(filename, "w") as o:
            json.dump(location_dictionary, o)
        logger.info(f"Full dump for location {location} written to {filename}")

    location = homely.get_location(locations[0].location_id)
    if args.stream:
        homely.run_socket_io(location, test_callback)



from .emitter import Emitter
from .parser import Parser, Packet
from .transports.polling import Polling

import gevent
import gevent.event
import gevent.queue
import json

import logging
logger = logging.getLogger(__name__)

class Client(Emitter):
    TRANSPORTS = {
        'polling': Polling
    }

    def __init__(self, scheme, hostname, port, path='/engine.io', transports=[], parser=None):
        super(Client, self).__init__()
        self.scheme = scheme
        self.hostname = hostname
        self.port = port
        self.path = path
        self.transports = [t for t in self.TRANSPORTS.keys() if t in transports] or self.TRANSPORTS.keys()
        self.parser = parser or Parser()

        self.state = 'closed'
        self.sid = None
        self.ping_interval = None
        self.ping_timeout = None
        self.pong_event = gevent.event.Event()
        self.send_queue = gevent.queue.JoinableQueue()
        self.transport = None
        self.ping_pong_loop = None
        self.flush_loop = None

    def open(self):
        self.state = 'opening'
        transport_name = self.transports[0]
        transport = self.create_transport(transport_name)
        
        transport.open()
        self.set_transport(transport)

    def close(self):
        if self.state not in ['opening', 'open']:
            return

        self.state = 'closing'
        self.send_queue.join()
        self.handle_close()

    def send(self, message, binary=False):
        self.send_packet(Packet(Packet.MESSAGE, message, binary))

    def create_transport(self, name):
        return self.TRANSPORTS[name](self, self.scheme, self.hostname, self.port, self.path, self.parser)

    def set_transport(self, transport):
        if self.transport:
            logger.debug("Clearing existing transport")
            self.transport.removeAllListeners()
    
        self.transport = transport
        self.transport.on('close', self.handle_close)
        self.transport.on('packet', self.handle_packet)
        self.transport.on('error', self.handle_error)

    def send_packet(self, packet):
        if self.state in ['closing', 'closed']:
            logger.warning("Trying to send a packet while state is: %s", self.state)
            return
        self.send_queue.put(packet)

    def loop_flush(self):
        while self.state in ['open', 'closing']:
            logger.debug("Waiting packets")
            self.send_queue.peek()
            logger.debug("Flushing packets")

            packets = []
            try:
                while True:
                    packet = self.send_queue.get_nowait()
                    packets.append(packet)
            except gevent.queue.Empty:
                pass

            self.transport.send(packets)
            for packet in packets:
                self.send_queue.task_done()

    def loop_ping_pong(self):
        while self.state in ['open', 'closing']:
            self.pong_event.clear()
            self.send_packet(Packet(Packet.PING))
            pong_received = self.pong_event.wait(timeout=self.ping_timeout/1000)
            if not pong_received:
                logger.warning("Pong timeout")
                self.handle_close()
                break
            gevent.sleep(self.ping_interval/1000)
    
    def start_loop(self, func, *args, **kwargs):
        def loop_stopped(g):
            logger.debug("Stop %s", func.__name__)
        g = gevent.spawn(func, *args, **kwargs)
        g.rawlink(loop_stopped)
        logger.debug("Start %s", func.__name__)
        return g

    def stop_loop(self, loop):
        if loop:
            loop.kill(block=False)

    def handle_open(self):
        self.state = 'open'
        self.emit('open')

    def handle_close(self):
        if self.state in ['opening', 'open', 'closing']:
            logger.debug("Closing client")
            self.state = 'closed'
            not_closed_by_transport = (self.state == 'closing')
            self.transport.close(send=not_closed_by_transport)
            self.sid = None
            self.stop_loop(self.ping_pong_loop)
            self.stop_loop(self.flush_loop)
            self.emit('close')

    def handle_handshake(self, handshake):
        self.sid = handshake['sid']
        self.ping_interval = handshake['pingInterval']
        self.ping_timeout = handshake['pingTimeout']
        self.handle_open()
        self.ping_pong_loop = self.start_loop(self.loop_ping_pong)
        self.flush_loop = self.start_loop(self.loop_flush)

    def handle_packet(self, packet):
        if self.state not in ['open', 'opening']:
            logger.warning("Packet received while state is: %s", self.state)
            return
    
        if packet.type == Packet.OPEN:
            handshake = json.loads(packet.data)
            self.handle_handshake(handshake)
        elif packet.type == Packet.CLOSE:
            self.transport.close(send=False)
        elif packet.type == Packet.PONG:
            self.pong_event.set()
        elif packet.type == Packet.MESSAGE:
            self.emit('message', packet.data)
        else:
            logger.warning("Invalid message type: %s", packet.type_string)

    def handle_error(self, error):
        logger.warning("Error occured: %s", error)
        self.emit('error', error)
        self.handle_close()

